/*
 * Copyright (c) 2000 Blue Mug, Inc.  All Rights Reserved.
 */

#include <assert.h>
#include <fcntl.h>
#include <poll.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <termios.h>
#include <time.h>
#include <unistd.h>

#include "eth.h"
#include "ethutil.h"
#include "options.h"
#include "serial.h"
#include "target.h"
#include "util.h"

#define MAX_COMMAND_LENGTH 128

#define ETHERNET_MTU 65536
#define SERIAL_MTU 4096

static const char TM_ETHERNET[] = "ethernet";
static const char TM_SERIAL[] = "serial";

/* ethernet helper functions */
static void target_get_mac(target_context_t *tc);
static void target_negotiate_mac(target_context_t *tc);
static void target_set_mac(target_context_t *tc, const unsigned char *mac);

static void target_write_ethernet(target_context_t *tc,
				  const void *data, size_t count);
static void target_write_serial(target_context_t *tc,
				const void *data, size_t count);

target_context_t *target_open(const char *port, const char *netif)
{
	target_context_t *tc;

	assert(port);

	tc = zmalloc(sizeof (target_context_t));
	tc->portfd = -1;
	tc->sockfd = -1;
	tc->medium = TM_SERIAL;
	tc->mtu = SERIAL_MTU;
	tc->write = target_write_serial;

	if ((tc->portfd = open_port(port)) < 0)
		goto failure;
	if (netif && (eth_open(tc, netif) < 0))
		goto failure;
	return tc;

failure:
	target_close(tc);
	return NULL;
}

int target_connect(target_context_t *tc)
{
	char buf [LINE_LENGTH];

	assert(tc);
	if (target_ping(tc) < 0)
		return -1;
	msg("target_connect: ping successful.\n");

	/* write 'version' command to get version */
	target_write_command(tc, "version");
	fprintf(msgout, "target: %s", target_gets(tc, buf, sizeof buf));
	fflush(msgout);
	target_confirm_response(tc);

	/* switch to ethernet if possible and desired */
	if (opt_ethernet)
		target_set_medium(tc, "ethernet");

	return 0;
}

void target_close(target_context_t *tc)
{
	assert(tc);
	close(tc->portfd);
	close(tc->sockfd);
	free(tc);
}

static void target_get_mac(target_context_t *tc)
{
	char buf [LINE_LENGTH];

	/* read MAC address from target */
	msg("getting target MAC\n");
	target_write_command(tc, "mac");
	target_gets(tc, buf, sizeof buf);
	target_confirm_response(tc);

	/* and parse it */
	msgf("target-provided remote MAC: %s", buf);
	if (parsemac(tc->remote_mac, buf))
		panicf("can't parse target-provided remote MAC: %s", buf);
	if (opt_verbose) {
		msg("target-provided remote MAC (parsed): ");
		printmac(msgout, tc->remote_mac);
		msg("\n");
	}
}

static void target_negotiate_mac(target_context_t *tc)
{
	unsigned char bogus_mac_1[6] = { 0, 0, 0, 0, 0, 0 };
	unsigned char bogus_mac_2[6] = { 255, 255, 255, 255, 255, 255 };

	assert(opt_ethernet);
	if (opt_remote_mac)
		target_set_mac(tc, opt_remote_mac_bytes);
	else
		target_get_mac(tc);

	/* check MAC validity */
	if (!memcmp(tc->remote_mac, bogus_mac_1, 6) ||
	    !memcmp(tc->remote_mac, bogus_mac_2, 6))
		panic("target MAC is invalid (need --remote-mac)");
}

static void target_set_mac(target_context_t *tc, const unsigned char *mac)
{
	char buf [CMDBUF_LENGTH];

	/* set MAC address on target */
	strcpy(buf, "mac ");
	sprintmac(buf + 4, mac);
	msgf("setting target MAC to `%s'\n", buf + 4);
	target_write_command(tc, buf);
	target_confirm_response(tc);
	
	/* update our local copy on success */
	memcpy(tc->remote_mac, mac, 6);
}

void target_set_medium(target_context_t *tc, const char *medium)
{
	char cmdbuf [CMDBUF_LENGTH];

	msgf("setting medium to %s\n", medium);

	/* if medium is Ethernet, negotiate remote MAC before switching */
	if (!strcmp(medium, TM_ETHERNET))
		target_negotiate_mac(tc);

	snprintf(cmdbuf, sizeof cmdbuf, "medium %s", medium);
	target_write_command(tc, cmdbuf);
	target_confirm_response(tc);
	if (!strcmp(medium, TM_ETHERNET)) {
		tc->mtu = ETHERNET_MTU;
		tc->write = target_write_ethernet;
	} else if (!strcmp(medium, TM_SERIAL)) {
		tc->mtu = SERIAL_MTU;
		tc->write = target_write_serial;
	} else
		panicf("unknown medium: %s", medium);
	tc->medium = medium;
}

void target_confirm_response(target_context_t *tc)
{
	if (target_get_response(tc))
		panic("target error, can't continue.");
}

int target_get_response(target_context_t *tc)
{
	char buf [LINE_LENGTH];

	assert(tc);
	switch (ok_or_ng(target_gets(tc, buf, sizeof buf))) {
	case 1:
		msg("target: +OK\n");
		break;
	case 2:
		msg("target: +GO\n");
		break;
	case 0:
		warn("target didn't respond with +OK or -NG\n");
		return -1;
	case -1:
		/* warning message is generated by target_gets() */
		return -1;
	default:
		assert(0);
		return -1;
	}
	return 0;
}

char *target_gets(target_context_t *tc, char *s, int size)
{
	unsigned char c;
	char *p = s;

	assert(tc);
	assert(s);
	assert(size >= 0);
	if (!size--)
		return s;
	while (size--) {
		xread(tc->portfd, &c, 1);
		if (c == '\r')
			++size;
		else
			*p++ = c;
		if (c == '\n')
			break;
	}
	*p = '\0';

	/* report error lines */
	if (ok_or_ng(s) < 0)
		warnf("from target: %s", s);
	return s;
}

void target_per_line(target_context_t *tc, iter_func_t *func, void *arg)
{
	char buf [LINE_LENGTH];

	assert(tc);
	assert(func);
	while (1) {
		target_gets(tc, buf, sizeof buf);	/* XXX handle error */
		msgf("read line from target: %s", buf);
		if (ok_or_ng(buf))
			break;
		if (func(buf, arg) < 0)
			break;
	}
}

/*
 * Try to put the target-side loader into batch mode; make sure it's
 * responding.  This is the only target function which times out.
 */
int target_ping(target_context_t *tc)
{
#define BUFLEN 127
	char buf [BUFLEN+1];	/* arbitrary size */
	struct pollfd pfd;
	char *ptr;
	time_t now, then;
	int ret;

	/* buf is always null-terminated */
	memset(buf, 0, sizeof buf);

	/* flush I/O on serial port */
	assert(tc);
	assert(tc->portfd >= 0);
	if (tcflush(tc->portfd, TCIOFLUSH) < 0)
		perror_xe("tcflush() on serial fd");

	/* couple-second timeout */
	now = time(NULL);
	then = now + 2;

	/* write a '~' (this puts the loader into batch mode) */
	memset(&pfd, 0, sizeof pfd);
	pfd.fd = tc->portfd;
	pfd.events = POLLOUT;
	ret = poll(&pfd, 1, 1000 * (int)(then - now));
	if (ret < 0)
		perror_xe("poll() on serial fd");
	if (!ret)
		goto nobody_home;
	buf[0] = '~';
	xwrite(pfd.fd, buf, 1);

	/* wait for "+OK" response */
	pfd.events = POLLIN;
	for (ptr = buf; ptr - buf < BUFLEN; /* nada */) {
		char *pos;
		now = time(NULL);
		if (now >= then)
			goto nobody_home;
		ret = poll(&pfd, 1, 1000 * (int)(then - now));
		if (ret < 0)
			perror_xe("poll() on serial fd");
		if (!ret)
			goto nobody_home;
		xread(pfd.fd, ptr++, 1);
		pos = strstr(buf, "+OK\r\n");
		if (pos && pos[5] == '\0')
			return 0;
	}
nobody_home:
	return -1;
}

void target_retry_command_and_data(target_context_t *tc,
				   const char *command,
				   const void *data, size_t count)
{
	int retries = 3;

	while (retries--) {
		msgf("trying command/data (retries = %d)\n", retries);
		target_write_command(tc, command);
		if (target_get_response(tc))
			continue;
		target_write_data(tc, data, count);
		if (target_get_response(tc) == 0)
			return;
	}
	panic("command/data write retry count exceeded.");
}

void target_write_command(target_context_t *tc, const char *command)
{
	char buf [MAX_COMMAND_LENGTH + 1];
	size_t len;

	assert(tc);
	assert(command);
	assert(strlen(command) < sizeof buf - 2);

	/* tack '\n' onto the end of the command */
	len = strlen(command);
	memcpy(buf, command, len);
	buf[len++] = '\n';
	buf[len] = '\0';
	msgf("writing command: %s", buf);
	assert(tc->write);
	tc->write(tc, buf, len);
}

void target_write_data(target_context_t *tc, const void *data, size_t count)
{
	assert(tc);
	assert(data);

	assert(tc->write);
	tc->write(tc, data, count);
	target_add_progress(tc, count);
}

static void target_write_ethernet(target_context_t *tc,
				  const void *data, size_t count)
{
	size_t remain, nsent;

	assert(tc->mtu == ETHERNET_MTU);
	assert(count <= tc->mtu);
	for (remain = count; remain; remain -= nsent) {
		char c;
		nsent = remain;
		if (nsent > ETH_DOWNLOAD_BLOCK)
			nsent = ETH_DOWNLOAD_BLOCK;
		eth_send_frame(tc, data, nsent, 1);
		/* wait for ack */
		xread(tc->portfd, &c, 1);
		assert(c == 'K');
		data += nsent;
	}
}

static void target_write_serial(target_context_t *tc,
				const void *data, size_t count)
{
	assert(tc->mtu == SERIAL_MTU);
	assert(count <= tc->mtu);
	xawrite(tc->portfd, data, count);
	tcdrain(tc->portfd);
}

void target_add_progress(target_context_t *tc, size_t progress)
{
	assert(tc);
	if (!tc->goal)
		return;
	tc->progress += progress;
	assert(tc->progress <= tc->goal);
	if (tc->progress >= tc->goal) {
		fprintf(msgout, "%s: completed 0x%08x (%u) bytes.          \n",
			tc->medium, tc->goal, tc->goal);
		tc->goal = 0;
		tc->progress = 0;
	} else {
		fprintf(msgout, "%s: 0x%08x (%u) bytes of %u...%c",
			tc->medium, tc->progress, tc->progress, tc->goal,
			opt_verbose ? '\n' : '\r');
	}
	fflush(msgout);
}

void target_set_goal(target_context_t *tc, size_t goal)
{
	assert(tc);
	tc->goal = goal;
	tc->progress = 0;
	target_add_progress(tc, 0);
}

int ok_or_ng(const char *s)
{
	assert(s);
	if (!strncmp(s, "+OK", 3))
		return 1;
	if (!strncmp(s, "+GO", 3))
		return 2;
	if (!strncmp(s, "-NG", 3))
		return -1;
	return 0;
}

