/*
 * Copyright (c) 2001 Atmark Techno, Inc.  All Rights Reserved.
 *
 * Commands to program the Fujitsu MBM29DL324BD flash chips on the
 * Armadillo board.  These are 32 Mbit "bottom boot"
 * (boot and param blocks at low end of address space) flash chips.
 */

#include <target/buffer.h>
#include <target/io.h>
#include <target/herrno.h>
#include <target/scan.h>

#include "flash.h"

#define FLASH_TIMEOUT 20000000

/* download buffer for the flash programmer */
#define DLBUFSIZE (8*1024)
unsigned char dlbuf [DLBUFSIZE];
word_t dlbufsize = DLBUFSIZE;

/* timeout for auto-boot, in seconds */
#define FL_WORD(addr) (*(volatile unsigned short*)(addr))

#define SR_MASK (SR7_WSMS|SR6_ESS|SR5_ES|SR4_PS|SR3_VPPS|SR2_PSS|SR1_BLLK)
#define SR_ERASE_ERR (SR5_ES|SR4_PS|SR3_VPPS|SR1_BLLK)
#define SR_BOTH_MASK (SM_MASK | (SM_MASK << 16))
#define SR_BOTH_WSMS (SR7_WSMS | (SR7_WSMS << 16))
#define SR_BOTH_ERASE_ERR (SR_ERASE_ERR | (SR_ERASE_ERR << 16))

/* handy flash functions */
static void flash_print_chip_error_string(unsigned char status);
static void flash_print_error_string(void);
static unsigned short flash_status_wait(addr_t addr, unsigned short value);
static int flash_status_full_check(addr_t addr, unsigned short value1, unsigned short value2);

static void flash_print_chip_error_string(unsigned char status)
{
	hprint("TimeOut");
}

/*
 * Come up with an error string representing the flash error; we can
 * have distinct errors for the two chips.  Get the status from
 * errdata, since this is used for error output.
 */
static void flash_print_error_string(void)
{
	hprint("lowbank:");
	flash_print_chip_error_string(errdata.word);
	hprint(" highbank:");
	flash_print_chip_error_string(errdata.word >> 16);
}

/*
 * Loop until both write state machines complete.
 */
static unsigned short flash_status_wait(addr_t addr, unsigned short value)
{
	unsigned short status;
	long timeout = FLASH_TIMEOUT;
	while (((status = (FL_WORD(addr))) != value) && timeout > 0) {
	  timeout--;
	}
	return status;
}

/*
 * Loop until the Write State machine is ready, then do a full error
 * check.  Clear status and leave the flash in Read Array mode; return
 * 0 for no error, -1 for error.
 */
static int flash_status_full_check(addr_t addr, unsigned short value1, unsigned short value2)
{
	unsigned short status1, status2;

	status1 = flash_status_wait(addr, value1);
	status2 = flash_status_wait(addr + 2, value2);
	if (status1 != value1 || status2 != value2) {
		errdata.word = status1 | status2 << 16;
		errfunc = &flash_print_error_string;
	}
	return (status1 != value1 || status2 != value2) ? -1 : 0;
}

/*
 * Program the contents of the download buffer to flash at the given
 * address.  Size is also specified; we shouldn't have to track usage
 * of the download buffer, since the host side can easily do that.
 *
 * We write words without checking status between each; we only go
 * through the full status check procedure once, when the full buffer
 * has been written.
 *
 * Alignment problems are errors here; we don't automatically correct
 * them because in the context of this command they signify bugs, and
 * we want to be extra careful when writing flash.
 */
static int flash_program_cmdfunc(int argc, char *argv[])
{
	addr_t addr;
	size_t size;
	unsigned short *ptr;
	int status = 0;
	addr_t base;

	if (argc != 3)
		return -H_EUSAGE;
	if (scan(*++argv, &addr)) return -H_EADDR;
	if (scan(*++argv, &size)) return -H_EADDR;
	if (addr & UNALIGNED_MASK) return -H_EALIGN;
	if (size & UNALIGNED_MASK) return -H_EALIGN;

	base = addr & 0xF0000000;

	size >>= BYTE_TO_WORD_SHIFT;
	for (ptr = (unsigned short*)dlbuf; size--; addr += sizeof(word_t)) {
		FL_WORD(base + (0x555 << 1)) = 0xAA;
		FL_WORD(base + (0x2AA << 1)) = 0x55;
		FL_WORD(base + (0x555 << 1)) = 0xA0;
		FL_WORD(addr) = *ptr;
		flash_status_wait(addr, *ptr);
		FL_WORD(base + (0x555 << 1)) = 0xAA;
		FL_WORD(base + (0x2AA << 1)) = 0x55;
		FL_WORD(base + (0x555 << 1)) = 0xA0;
		FL_WORD(addr + 2) = *(ptr + 1);
		flash_status_wait(addr + 2, *(ptr + 1));
		status |= flash_status_full_check(addr, *ptr, *(ptr + 1));
		ptr += 2;
	}
	return status;
}

const command_t flash_program_command =
	{ "program", "<addr> <size>",
	  "program flash from download buffer",
	  &flash_program_cmdfunc };

/*
 * Erase a flash block.  Single argument is the first address in the
 * block (actually, it can be anywhere in the block, but the host-side
 * hermit downloader gives the first address).
 */
static int flash_erase_cmdfunc(int argc, char *argv[])
{
	addr_t addr;
	addr_t base;

	if (argc != 2)
		return -H_EUSAGE;
	if (scan(*++argv, &addr)) return -H_EADDR;
	addr &= ~(addr_t)UNALIGNED_MASK;	/* word align */

	base = addr & 0xF0000000;

	FL_WORD(base + (0x555 << 1)) = 0xAA;
	FL_WORD(base + (0x2AA << 1)) = 0x55;
	FL_WORD(base + (0x555 << 1)) = 0x80;
	FL_WORD(base + (0x555 << 1)) = 0xAA;
	FL_WORD(base + (0x2AA << 1)) = 0x55;
	FL_WORD(addr) = 0x30;
	return flash_status_full_check(addr, 0xFFFF, 0xFFFF);
}

const command_t flash_erase_command =
	{ "erase", "<addr>", "erase flash block", &flash_erase_cmdfunc };
